# coding: utf-8
"""
Description:
    This model is used to store translations. To do so, we built the following structure:
    - Field --> Item --> Translation <-- Language
    - Field and Language instances must be inserted manually
    - Item and Translation instances are both created and deleted automatically, using "signals.py" and "CASCADE"
    A few things to note are:
        - We can traceback a Translation to its application and model (using ContentType)
        - If a model creates other model instances with signals, we block its "bulk_create" method with a custom manager
        - As a reminder, bulk_create does not return PK, so we can only use it on the "last" table of any chain reaction
Abstract Models:
    TranslatedModel: Abstract model to be used as parent for any model that requires translation
Models:
    Field: Lookup table that contains the list of fields eligible for translation.
    Item: Content table that stores the actual item that must be translated (object + field).
    Language: Lookup table that contains the list of available languages.
    Translation: Content table that stores all the available translations
"""


# --------------------------------------------------------------------------------
# > Imports
# --------------------------------------------------------------------------------
# Built-in

# Django
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.validators import MinLengthValidator
from django.db import models

# Third-party

# Local
from .fields import ForeignKeyCascade, NotEmptyCharField
from .managers import NoBulkCreateManager


# --------------------------------------------------------------------------------
# > Abstract Models
# --------------------------------------------------------------------------------
class TranslatedModel(models.Model):
    """
    Abstract model used as a template for any model that requires translation. It provides:
        - A basic "meta_info" field
        - A manager that prevents the "objects.bulk_create" method
        - A basic "__str__" method
        - methods to easily get translation info from the instance
    """
    # ----------------------------------------
    # Constants
    # ----------------------------------------
    TRANSLATED = True

    # ----------------------------------------
    # Fields
    # ----------------------------------------
    meta_info = models.CharField(
        blank=False,
        help_text="Information on the instance (not visible in the frontend)",
        max_length=100,
        null=False,
        verbose_name="Meta Information",
    )

    # ----------------------------------------
    # Custom Managers
    # ----------------------------------------
    objects = NoBulkCreateManager()

    # ----------------------------------------
    # META, str, save, get_absolute_url
    # ----------------------------------------
    class Meta:
        """Metadata to configure our model in the database"""
        abstract = True

    def __str__(self):
        """Returns the meta information"""
        return self.meta_info

    # ----------------------------------------
    # Custom Methods
    # ----------------------------------------
    def get_content_type_instance(self):
        """Returns the ContentType instance of our object"""
        app_label = self._meta.app_label
        model = self.__class__.__name__.lower()
        content_type = ContentType.objects.get(app_label=app_label, model=model)
        return content_type

    def get_translated_fields(self):
        """Returns a QuerySet of all the Field instances associated with the Model of our instance"""
        content_type = self.get_content_type_instance()
        fields = Field.objects.filter(content_type=content_type)
        return fields

    def get_translated_items(self):
        """Returns a QuerySet of all the Item instances associated with our instance"""
        fields = self.get_translated_fields()
        items = Item.objects.filter(object_id=self.id, field__in=fields)
        return items

    def get_translations(self):
        """Returns a QuerySet of all the Translation instances associated with our instance"""
        items = self.get_translated_items()
        translations = Translation.objects.filter(item__in=items)
        return translations


# --------------------------------------------------------------------------------
# > Models
# --------------------------------------------------------------------------------
class Field(models.Model):
    """
    Lookup table that contains the list of fields eligible for translation.
    Each Field is linked to a ContentType (ie an application's model).
    Field sends a signal on create that generates "Item" instances.
    """

    # ----------------------------------------
    # Fields
    # ----------------------------------------
    content_type = ForeignKeyCascade(
        ContentType,
        verbose_name="Content Type"
    )
    name = NotEmptyCharField(
        max_length=100,
        verbose_name="Field name"
    )

    # ----------------------------------------
    # Custom Managers
    # ----------------------------------------
    objects = NoBulkCreateManager()

    # ----------------------------------------
    # META, str, save, get_absolute_url
    # ----------------------------------------
    class Meta:
        """Metadata to configure our model in the database"""
        db_table = "ddt_fields"
        indexes = []
        ordering = [
            "content_type",
            "name"
        ]
        unique_together = [
            ['content_type', 'name'],
        ]
        verbose_name = "Field"
        verbose_name_plural = "Fields"

    def __str__(self):
        """Returns 'application.model.field' as a string"""
        app = self.content_type.app_label
        model = self.content_type
        return "{}.{}.{}".format(app, model, self.name)

    # ----------------------------------------
    # Custom Properties
    # ----------------------------------------
    def count_items(self):
        """Returns the amount of Item instances attached to this field"""
        return Item.objects.filter(field=self).count()
    count_items.short_description = "Items"

    def count_missing_translations(self):
        """Returns the amount of empty Translation instances attached to this field"""
        return Translation.objects.filter(item__field=self).filter(text="").count()
    count_missing_translations.short_description = "Missing Translations"

    def get_app_name(self):
        """Returns the application name as a string"""
        return self.content_type.app_label
    get_app_name.short_description = "Application"


class Item(models.Model):
    """
    Content table that stores the actual item that must be translated (object + field).
    Since it contains ID from different tables, we used a GenericForeignKey.
    Item sends a signal on create that generates "Translation" instances.
    """

    # ----------------------------------------
    # Fields
    # ----------------------------------------
    field = ForeignKeyCascade(
        "Field",
        verbose_name="Field"
    )
    object_id = models.PositiveIntegerField(
        null=False,
        db_index=True,
        verbose_name="Object ID"
    )
    content_type = ForeignKeyCascade(
        ContentType,
        verbose_name="Content Type"
    )
    of = GenericForeignKey('content_type', 'object_id')

    # ----------------------------------------
    # Custom Managers
    # ----------------------------------------
    objects = NoBulkCreateManager()

    # ----------------------------------------
    # META, str, save, get_absolute_url
    # ----------------------------------------
    class Meta:
        """Metadata to configure our model in the database"""
        db_table = "ddt_items"
        indexes = []
        ordering = []
        unique_together = [
            ['field', 'object_id'],
        ]
        verbose_name = "Item"
        verbose_name_plural = "Items"

    def __str__(self):
        """Returns the 'application.model.field.object_id' as a string"""
        return "{}.{}".format(self.field, self.object_id)

    # ----------------------------------------
    # Custom Properties
    # ----------------------------------------
    def count_missing_translations(self):
        """Returns the amount of empty Translation instances attached to this item"""
        return Translation.objects.filter(item=self).filter(text="").count()
    count_missing_translations.short_description = "Missing Translations"


class Language(models.Model):
    """
    Lookup table that contains the list of available languages.
    Used for translating the database and the frontend.
    Language sends a signal on create that generates "Translation" instances.
    """

    # ----------------------------------------
    # Fields
    # ----------------------------------------
    name = NotEmptyCharField(
        max_length=100,
        unique=True,
        verbose_name="Name"
    )
    iso2 = NotEmptyCharField(
        max_length=2,
        unique=True,
        validators=[MinLengthValidator(2)],
        verbose_name="ISO2"
    )
    iso3 = NotEmptyCharField(
        max_length=3,
        unique=True,
        validators=[MinLengthValidator(3)],
        verbose_name="ISO3"
    )
    django_language_name = NotEmptyCharField(
        help_text="Should match a value from settings.LANGUAGES, such as 'fr-FR' or 'en-US'",
        max_length=5,
        unique=True,
        verbose_name="Django language name"
    )

    # ----------------------------------------
    # Custom Managers
    # ----------------------------------------
    objects = NoBulkCreateManager()

    # ----------------------------------------
    # META, str, save, get_absolute_url
    # ----------------------------------------
    class Meta:
        """Metadata to configure our model in the database"""
        db_table = "ddt_languages"
        indexes = []
        ordering = ["name"]
        unique_together = []
        verbose_name = "Language"
        verbose_name_plural = "Languages"

    def __str__(self):
        """Returns the language name"""
        return self.name

    def save(self, *args, **kwargs):
        """Overrides the save method to 'upper' ISO2 and ISO3"""
        self.iso2 = self.iso2.upper()
        self.iso3 = self.iso3.upper()
        super().save(*args, **kwargs)

    # ----------------------------------------
    # Custom Methods
    # ----------------------------------------
    def count_missing_translations(self):
        """Returns the amount of empty Translation instances attached to this language"""
        return Translation.objects.filter(language=self).filter(text="").count()
    count_missing_translations.short_description = "Missing Translations"


class Translation(models.Model):
    """
    Content table that stores all the available translations.
    Each entry contains a language and an item (both ForeignKey).
    """

    # ----------------------------------------
    # Fields
    # ----------------------------------------
    language = ForeignKeyCascade(
        "Language",
        verbose_name="Language"
    )
    item = ForeignKeyCascade(
        "Item",
        verbose_name="Item"
    )
    text = models.TextField(
        default="",
        null=False,
        verbose_name="Translated text"
    )

    # ----------------------------------------
    # META, str, save, get_absolute_url
    # ----------------------------------------
    class Meta:
        db_table = "ddt_translations"
        indexes = [
            models.Index(
                fields=["item", "language"],
                name="idx_item_lang",
            )
        ]
        ordering = []
        unique_together = [
            ["language", "item"],
        ]
        verbose_name = "Translation"
        verbose_name_plural = "Translations"

    def __str__(self):
        """Returns a string with the 'item' and the 'language'"""
        return "{} ({})".format(self.item, self.language)

    # ----------------------------------------
    # Custom Properties
    # ----------------------------------------
    def truncated_text(self):
        """Returns the first 20 characters from self.text"""
        try:
            return self.text[:20]
        except IndexError:
            return self.text
    truncated_text.short_description = "Translated Text"



# coding: utf-8
"""
Description:
    This file allows us to personalize the admin interface.
    We can change the way tables and fields are displayed.
Helpers:
    ApplicationFilter: Allows filtering using the application name, on Field, Item, and Translation
    ContentTypeDropdown: Allows us to override how Content Type are displayed in a dropdown
Inlines:
    TranslationInline: Displays and allows editing of existing translations of an Item
Abstract Admins:
    TranslatedAdmin: ModelAdmin to use as parent for any model that has fields to translate
Admins:
    FieldAdmin: Customizes the Field model in the administration interface
    ItemAdmin: Customizes the Item model in the administration interface
    LanguageAdmin: Customizes the Language model in the administration interface
    TranslationAdmin: Customizes the Translation model in the administration interface
"""


# --------------------------------------------------------------------------------
# > Imports
# --------------------------------------------------------------------------------
# Built-in

# Django
from django.contrib.admin.utils import flatten_fieldsets
from django.contrib import admin
from django.contrib.contenttypes.models import ContentType
from django.forms import ModelChoiceField

# Third-party

# Local
from .forms import DynamicTranslationForm, create_translation_fieldname
from .models import Field, Item, Language, Translation
from .signals import translated_models


# --------------------------------------------------------------------------------
# > Helpers
# --------------------------------------------------------------------------------
class ApplicationFilter(admin.SimpleListFilter):
    """Allows filtering using the application name on Field, Item, and Translation"""

    title = 'Application'
    parameter_name = 'Application'

    def lookups(self, request, model_admin):
        """Returns a list of tuples (value, label) that will be the filter options"""
        fields = Field.objects.all()
        results = {field.get_app_name() for field in fields}
        results = [(value, value.capitalize()) for value in results]
        return results

    def queryset(self, request, queryset):
        """Defines the results of the filter query"""
        # Getting the model name from URL
        url = request.build_absolute_uri("?")
        if url[-1] == "/":
            url = url[:-1]
        model_name = url.split("/")[-1]
        # Changing the query based on our model
        if self.value() is not None:
            if model_name in {"field", "item"}:
                return queryset.filter(content_type__app_label=self.value())
            elif model_name == "translation":
                return queryset.filter(item__content_type__app_label=self.value())
        else:
            return queryset


class ContentTypeDropdown(ModelChoiceField):
    """Allows us to override how Content Type are displayed in a dropdown"""

    def label_from_instance(self, obj):
        """Returns a string that specifies both the APP and the MODEL name"""
        app = obj.app_label.upper()
        model = obj.model.upper()
        return "({}) {}".format(app, model)


# --------------------------------------------------------------------------------
# > Inlines
# --------------------------------------------------------------------------------
class TranslationInline(admin.TabularInline):
    """Inline model to display the Translation instances on an Item (update) form"""

    # ----------------------------------------
    # Config
    # ----------------------------------------
    model = Translation
    can_delete = False
    extra = 0
    max_num = 0

    # ----------------------------------------
    # Fields
    # ----------------------------------------
    readonly_fields = ("language",)
    fieldsets = (
        [
            "",
            {
                "fields": [
                    "language",
                    "text"
                ]
            },
        ],
    )


# --------------------------------------------------------------------------------
# > Abstract Admins
# --------------------------------------------------------------------------------
class TranslatedAdmin(admin.ModelAdmin):
    """
    ModelAdmin to use as parent for any model that has fields to translate
    It comes with the "DynamicTranslationForm" and custom methods to display its fields
    """

    # ----------------------------------------
    # Config
    # ----------------------------------------
    form = DynamicTranslationForm

    # ----------------------------------------
    # Detail View
    # ----------------------------------------
    fieldsets = []

    # ----------------------------------------
    # Custom Methods
    # ----------------------------------------
    def get_form(self, request, obj=None, **kwargs):
        """Required for get_fieldsets"""
        kwargs['fields'] = flatten_fieldsets(self.fieldsets)
        return super().get_form(request, obj, **kwargs)

    def get_fieldsets(self, request, obj=None):
        """
        Allows us to display the field dynamically created by "DynamicTranslationForm"
        The fieldnames in "DynamicTranslationForm" and this function must be identical
        In other words:
        - "DynamicTranslationForm" creates the fields
        - This function adds fields with the same name to the fieldsets
        - As a result, the fields now appear on the form
        """
        fieldsets = self.fieldsets.copy()
        # Get current Item
        url = request.build_absolute_uri("?")
        if url.endswith("/change/"):
            url = url[:-8]
            object_id = url.split("/")[-1]
            obj = self.model.objects.get(pk=object_id)
            # Create a field for each translation associated with our object
            fields = []
            translations = obj.get_translations()
            for translation in translations:
                fieldname = create_translation_fieldname(translation)
                fields.append(fieldname)
            # Add a fieldset with our fields
            fieldsets.append(['TRANSLATIONS', {'fields': fields}])
        return fieldsets


# --------------------------------------------------------------------------------
# > Admins
# --------------------------------------------------------------------------------
@admin.register(Field)
class FieldAdmin(admin.ModelAdmin):
    """Customizes the Field model in the administration interface"""

    # ----------------------------------------
    # List view
    # ----------------------------------------
    list_display = [
        "id",
        "get_app_name",
        "content_type",
        "name",
        "count_items",
        "count_missing_translations"
    ]
    list_display_links = [
        "id",
        "name",
    ]
    list_editable = []
    list_filter = [
        ApplicationFilter,
        ("content_type", admin.RelatedOnlyFieldListFilter)
    ]
    ordering = ["-id"]
    search_fields = ["name"]
    sortable_by = [
        "id",
        "content_type",
        "name",
    ]

    # ----------------------------------------
    # Detail view
    # ----------------------------------------
    autocomplete_fields = []
    readonly_fields = [
        "id",
        "count_items",
        "count_missing_translations",
    ]
    radio_fields = {}
    raw_id_fields = []
    fieldsets = [
        [
            "CORE INFORMATION", {
                "fields": ["id"],
            }
        ],
        [
            "DYNAMIC STATISTICS", {
                "fields": [
                    "count_items",
                    "count_missing_translations"
                ],
            }
        ],
        [
            "CONTENT", {
                "fields": [
                    "content_type",
                    "name"
                ],
            }
        ],
    ]

    # ----------------------------------------
    # Custom Methods
    # ----------------------------------------
    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        """Allows us to override how Content Type are displayed in a dropdown"""
        if db_field.name == "content_type":
            types = [ContentType.objects.get_for_model(model).pk for model in translated_models]
            kwargs['queryset'] = ContentType.objects.filter(pk__in=types)
            return ContentTypeDropdown(**kwargs)
        return super().formfield_for_foreignkey(db_field, request, **kwargs)


@admin.register(Item)
class ItemAdmin(admin.ModelAdmin):
    """
    Customizes the Item model in the administration interface
    You should not create, delete, nor edit instances from the admin
    This table is entirely generated/updated through 'signals.py' and CASCADE relationships
    """

    # ----------------------------------------
    # List view
    # ----------------------------------------
    list_display = [
        "id",
        "field",
        "object_id",
        "count_missing_translations"
    ]
    list_display_links = None
    list_editable = []
    list_filter = [
        ApplicationFilter,
        ("content_type", admin.RelatedOnlyFieldListFilter),
        "field"
    ]
    ordering = ["-id"]
    search_fields = ["object_id"]
    sortable_by = [
        "id",
        "field",
        "object_id",
    ]

    # ----------------------------------------
    # Detail view
    # ----------------------------------------
    autocomplete_fields = []
    readonly_fields = [  # All fields are read-only, so no edit possible
        "id",
        "field",
        "object_id",
        "count_missing_translations",
    ]
    radio_fields = {}
    raw_id_fields = []
    fieldsets = [
        [
            "CORE INFORMATION", {
                "fields": [
                    "id",
                    "field",
                    "object_id",
                ],
            }
        ],
        [
            "DYNAMIC STATISTICS", {
                "fields": ["count_missing_translations"],
            }
        ],
    ]
    inlines = [TranslationInline]


@admin.register(Language)
class LanguageAdmin(admin.ModelAdmin):
    """Customizes the Language model in the administration interface"""

    # ----------------------------------------
    # List view
    # ----------------------------------------
    list_display = [
        "id",
        "name",
        "iso2",
        "iso3",
        "django_language_name",
        "count_missing_translations"
    ]
    list_display_links = [
        "id",
        "name",
    ]
    list_editable = []
    list_filter = []
    ordering = ["-id"]
    search_fields = [
        "name",
        "iso2",
        "iso3"
        "django_language_name",
    ]
    sortable_by = [
        "id",
        "name",
        "iso2",
        "iso3",
        "django_language_name",
    ]

    # ----------------------------------------
    # Detail view
    # ----------------------------------------
    autocomplete_fields = []
    readonly_fields = [
        "id",
        "count_missing_translations",
    ]
    radio_fields = {}
    raw_id_fields = []
    fieldsets = [
        [
            "CORE INFORMATION", {
                "fields": ["id"],
            }
        ],
        [
            "DYNAMIC STATISTICS", {
                "fields": ["count_missing_translations"],
            }
        ],
        [
            "CONTENT", {
                "fields": [
                    "name",
                    "iso2",
                    "iso3",
                    "django_language_name",
                ],
            }
        ],
    ]


# Remove the comment below to access Item within the admin
# Only for debugging purposes
@admin.register(Translation)
class TranslationAdmin(admin.ModelAdmin):
    """
    Customizes the Translation model in the administration interface
    Instances sould not be created nor deleted through the admin interface
    Creation and deletion are entirely handled by 'signals.py' and CASCADE relationships
    It is however possible to edit the "text" field through the detail view
    """

    # ----------------------------------------
    # List view
    # ----------------------------------------
    # actions = None
    list_display = [
        "id",
        "language",
        "item",
        "truncated_text"
    ]
    list_display_links = ["id", "truncated_text"]
    list_editable = []
    list_filter = [
        ApplicationFilter,
        ("item__content_type", admin.RelatedOnlyFieldListFilter),
        "item__field",
        "item",
        "language",
    ]
    ordering = ["-id"]
    search_fields = ["truncated_text"]
    sortable_by = [
        "id",
        "language",
        "item",
    ]

    # ----------------------------------------
    # Detail view
    # ----------------------------------------
    autocomplete_fields = []
    readonly_fields = [
        "id",
        "language",
        "item",
    ]
    radio_fields = {}
    raw_id_fields = []
    fieldsets = [
        [
            "CORE INFORMATION", {
                "fields": [
                    "id",
                    "language",
                    "item",
                ],
            }
        ],
        [
            "CONTENT", {
                "fields": ["text"],
            }
        ],
    ]
